import os

import torch
import numpy as np

from tqdm import tqdm
from torch.nn.utils.rnn import pad_sequence
from transformers import AutoModel, AutoTokenizer

from ..networks import PolicyNetwork, ValueNetwork
from .memory import Memory

class Agent:
    def __init__(
        self,
        action_dims: int = None,
        gamma: float = 0.99,
        epochs: int = 10,
        gae_lambda: float =0.95,
        policy_clip: float = 0.1,
        entropy: float = 0.001,
        batch_size: int = 64,
        encoder_name: str = 'bert-base-uncased',
        tokenizer_name: str = 'bert-base-uncased',
        actor_ridge: float = 0,
        actor_coefficient: float = 1.,
        actor_clip_grad_norm: float = None,
        critic_ridge: float = 0,
        critic_coefficient: float = 1.,
        critic_clip_grad_norm: float = None,
        name: str = 'alector',
        verbose: bool = False,
        load: bool = False,
        chkpt_dir: str = 'chkpts',
        device: str = 'cpu',
        **kwargs
        ):

        actor_hparams = {
            **kwargs.get("actor_hparams", {})
        }
        critic_hparams = {
            **kwargs.get("critic_hparams", {})
        }
        encoder_hparams = {
            **kwargs.get("encoder_hparams", {})
        }
        tokenizer_hparams = {
            **kwargs.get("tokenizer_hparams", {})
        }
        
        self.gamma = gamma
        self.epochs = epochs
        self.gae_lambda = gae_lambda
        self.policy_clip = policy_clip

        self.actor_coef = actor_coefficient
        self.actor_ridge_lambda = actor_ridge

        self.critic_coef = critic_coefficient
        self.critic_ridge_lambda = critic_ridge

        self.entropy_coef = entropy

        self.name = name
        self.verbose = verbose
        self.path = os.path.join(
            chkpt_dir,
            'ppo'
        )
        os.makedirs(self.path, exist_ok=True)
 
        self.device = device

        self.encoder = AutoModel.from_pretrained(
            pretrained_model_name_or_path=encoder_name,
            **encoder_hparams
        )
        self.encoder.eval()
        self.encoder.to(device)
        embed_dims = self.encoder.config.hidden_size

        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=tokenizer_name,
            **tokenizer_hparams
        )
        vocab_size = self.tokenizer.vocab_size

        output_dims = action_dims or vocab_size

        self.actor = PolicyNetwork(
            input_dims=embed_dims,
            output_dims=output_dims,
            device=device,
            **actor_hparams
        )
        self.actor_clip_grad_norm = actor_clip_grad_norm

        self.critic = ValueNetwork(
            input_dims=embed_dims,
            device=device,
            **critic_hparams
        )
        self.critic_clip_grad_norm = critic_clip_grad_norm

        if load:
            if os.path.exists(os.path.join(self.path, f"{self.name}.pt")):
                self.load()
            else:
                if self.verbose:
                    print("No possible file to load from found. Initialising new agent")

        self.memory = Memory(
            batch_size=batch_size
        )

        if self.verbose:
            if load:
                print(f"agent successfully loaded from {self.path}/{self.name}")
            actor_params = sum(p.numel() for p in self.actor.parameters())
            critic_params = sum(p.numel() for p in self.critic.parameters())
            print(f"encoder ready on {self.device}. using {embed_dims} embedding dims")
            print(f"actor ready on {self.device}. total params: {actor_params}")
            print(f"critic ready on {self.device}. total params: {critic_params}")
            print(f"{self.name} initialised with {actor_params+critic_params} params.")
            print(f"tokenizer using {vocab_size} different tokens")

    @torch.no_grad()
    def choose_action(self, state):
        encoded_state = self.tokenizer(
            state,
            padding=True,
            truncation=True,
            return_tensors='pt'
        ).to(self.device)
        encoder_output = self.encoder(**encoded_state)
        embeddings = encoder_output['last_hidden_state']

        dist = self.actor(embeddings)
        value = self.critic(embeddings)
        action = dist.sample()

        probs = dist.log_prob(action)

        return action, probs, value, embeddings

    def save(self):
        to_save = dict(
            actor = dict(
                network = self.actor.state_dict(),
                optimizer = self.actor.optimizer.state_dict()
            ),
            critic = dict(
                network = self.critic.state_dict(),
                optimizer = self.critic.optimizer.state_dict()
            )
        )
        torch.save(
            to_save,
            os.path.join(
                self.path,
                f"{self.name}.pt"
            )
        )

    def load(self):
        chkpt = torch.load(
            os.path.join(
                self.path,
                f"{self.name}.pt"
            ),
            map_location=self.device,
            weights_only=True
        )
        self.actor.load_state_dict(
            chkpt['actor']['network']
        )
        self.actor.optimizer.load_state_dict(
            chkpt['actor']['optimizer']
        )
        self.critic.load_state_dict(
            chkpt['critic']['network']
        )
        self.critic.optimizer.load_state_dict(
            chkpt['critic']['optimizer']
        )

    def learn(self):
        actor_loss_arr = []
        critic_loss_arr = []
        entropy_loss_arr = []
        total_loss_arr = []
        def calculate_total_loss(actor_loss, critic_loss, entropy):
            actor_contrib = self.actor_coef*actor_loss
            critic_contrib = self.critic_coef*critic_loss
            entropy_contrib = self.entropy_coef*entropy
            loss = actor_contrib + critic_contrib - entropy_contrib
            return loss

        for _ in tqdm(range(self.epochs), disable=not self.verbose, desc="Learning Epoch", leave=False):
            state_arr, action_arr, old_probs_arr, vals_arr, reward_arr, dones_arr, batches = self.memory.recall()

            advantage = np.zeros(len(reward_arr), dtype=np.float32)
            for t in range(len(reward_arr)-1):
                discount = 1
                a_t = 0
                for k in range(t, len(reward_arr)-1):
                    a_t += discount*(reward_arr[k]+self.gamma*vals_arr[k+1]\
                            *(1-int(dones_arr[k]))-vals_arr[k])
                    discount *= self.gamma*self.gae_lambda
                advantage[t] = a_t

            for batch in batches:
                self.actor.optimizer.zero_grad()
                self.critic.optimizer.zero_grad()

                states = pad_sequence(
                    [state_arr[idx].squeeze(0) for idx in batch],
                    batch_first=True
                ).to(self.device)

                dist = self.actor(states)
                value = self.critic(states)
                del states

                entropy = dist.entropy().mean()
                entropy_loss_arr.append(entropy.item())

                actions = torch.tensor(action_arr[batch]).to(self.device)
                new_probs = dist.log_prob(actions)
                del dist, actions

                old_probs = torch.tensor(old_probs_arr[batch]).to(self.device)
                prob_ratio = (new_probs-old_probs).exp()
                del new_probs, old_probs 

                batch_advantages = torch.tensor(
                    advantage[batch]
                ).to(self.device)

                weighted_probs = batch_advantages*prob_ratio
                weighted_clipped_probs = torch.clamp(
                    prob_ratio,
                    min=1-self.policy_clip,
                    max=1+self.policy_clip
                )*batch_advantages
                del prob_ratio

                actor_loss = torch.min(
                    weighted_probs,
                    weighted_clipped_probs
                ).mean()
                del weighted_probs, weighted_clipped_probs

                actor_ridge = sum(p.pow(2).sum() for p in self.actor.parameters())
                actor_loss += self.actor_ridge_lambda*actor_ridge
                actor_loss_arr.append(actor_loss.item())
                del actor_ridge

                batch_values = torch.tensor(vals_arr[batch]).to(self.device)
                returns = batch_advantages+batch_values
                del batch_values, batch_advantages

                critic_loss = ((value-returns)**2).mean()
                critic_ridge = sum(p.pow(2).sum() for p in self.critic.parameters())
                critic_loss += self.critic_ridge_lambda*critic_ridge
                critic_loss_arr.append(critic_loss.item())
                del returns, critic_ridge

                total_loss = calculate_total_loss(actor_loss,critic_loss, entropy)
                total_loss.backward()
                total_loss_arr.append(total_loss.item())

                if self.actor_clip_grad_norm is not None:
                    torch.nn.utils.clip_grad_norm_(
                        self.actor.parameters(),
                        max_norm=self.actor_clip_grad_norm
                    )

                if self.critic_clip_grad_norm is not None:
                    torch.nn.utils.clip_grad_norm_(
                        self.critic.parameters(),
                        max_norm=self.critic_clip_grad_norm
                    )

                self.actor.optimizer.step()
                self.critic.optimizer.step()

                torch.cuda.empty_cache()
        self.memory.clear()
        return dict(
            actor_loss = np.mean(actor_loss_arr),
            critic_loss = np.mean(critic_loss_arr),
            entropy = np.mean(entropy_loss_arr),
            total_loss = np.mean(total_loss_arr)
        )



import torch.nn as nn

def generate_layers(
    input_dims: int,
    output_dims: int,
    layers: list,
    activation_fn: str = 'none'
):

    activation_fn = activation_fn.lower()


    generated_layers = [
        nn.Linear(input_dims, layers[0]),
    ]

    for i in range(len(layers)-1):
        if activation_fn == 'relu':
            generated_layers.append(
                nn.ReLU()
            )
        elif activation_fn == 'leakyrelu':
            generated_layers.append(
                nn.LeakyReLU()
            )
        elif activation_fn == 'tanh':
            generated_layers.append(
                nn.Tanh()
            )

        generated_layers.append(
            nn.Linear(layers[i], layers[i+1])
        )

    return generated_layers




import warnings
import torch
import torch.nn as nn

from .transformer import TransformerBlock, AttentionGate

from ..optimizers import Adam, UPGD

from ..utils import generate_layers

class ValueNetwork(nn.Module):
    def __init__(
        self,
        input_dims,
        transformer_layers: int = 2,
        optimizer_name: str = 'adam',
        device='cpu',
        **kwargs
        ):
        gate_hparams = {
            **kwargs.get("gate", {})
        }
        network_defaults = dict(
            layers=None,
            activation_fn=None
        )
        network_architecture = {
            **network_defaults,
            **kwargs.get("network", {})
        }
        transformer_hparams = {
            **kwargs.get("transformer", {})
        }
        super(ValueNetwork, self).__init__()

        self.transformer = nn.ModuleList(
            [
                TransformerBlock(
                    embed_dims=input_dims,
                    **transformer_hparams,
                )
                for _ in range(transformer_layers)
            ]
        )

        self.gate = AttentionGate(
            embed_dims=input_dims,
            **gate_hparams
        )

        if network_architecture['layers'] is not None:
            layers = generate_layers(
                input_dims=input_dims,
                output_dims=1,
                **network_architecture
            )
            self.critic = nn.ModuleList(layers)
        else:
            self.critic = nn.ModuleList(
                [
                    nn.Linear(input_dims, 1)
                ]
            )

        self.to(device)

        ## OPTIMIZER
        optimizer = kwargs.get("optimizer")
        if isinstance(optimizer, type) and issubclass(optimizer, Optimizer):
            optimizer_hparams = {
                **kwargs.get("optimizer_hparams", {})
            }
            self.optimizer = optimizer(
                self.parameters(),
                **optimizer_hparams
            )
        else: 
            ####### TO BE REMOVED DO NOT USE THIS
            warnings.warn(
                "Passing 'optimizer' as a dict to the actor/critic is deprecated. "
                "Please pass a pytorch.optim class in 'optimizer' and hyperparameters in 'optimizer_hparams' dict.",
                Warning
            )
            optimizer_hparams = {k: v for k, v in optimizer.items()}
            if 'upgd' in optimizer_name.lower():
                self.optimizer = UPGD(
                    params=self.parameters(),
                    **optimizer_hparams
                )
            else:
                self.optimizer = Adam(
                    params=self.parameters(),
                    **optimizer_hparams
                )
            #####################################

    def forward(self, state, mask=None):
        out = state
        for layer in self.transformer:
            out = layer(out, out, out, mask)
        out = self.gate(out)
        for layer in self.critic:
            out = layer(out)
        value = out
        return value



import torch
import torch.optim as optim

class UPGD(optim.Optimizer):
    """
        LICENSED UNDER MIT BY Mohamed Elsayed
        Paper: https://openreview.net/forum?id=sKPzAXoylB
        Code: https://github.com/mohmdelsayed/upgd
        LICENSED UNDER MIT BY Mohamed Elsayed

        Utility-based Perturbed Gradient Descent (UPGD) is an optimization method designed for
        continual learning in neural networks. It addresses two major issues of continual learning:
        catastrophic forgetting and loss of plasticity.

        The UPGD method uses a utility-based gating mechanism to modify the weights:
        - Weights with high utility are not sugnificantly updated, preventing them from being forgotten.
        - Weights with low utility are perturbed, allowing for plasticity in learning new tasks.

        The update rule also incorporates noise to further encourage plasticity, similar to Perturbed Gradient Descent (PGD).
        This method can be used in both standard and streaming learning setups.
    """
    def __init__(
        self,
        params,
        lr=1e-5,
        decay=0.001,
        beta=0.999,
        sigma=0.001
    ):

        hparams = dict(
            lr = lr,
            weight_decay = decay,
            beta_utility = beta,
            sigma = sigma,
        )
        
        super(UPGD, self).__init__(params, hparams)

    def step(self):
        global_max_util = torch.tensor(-torch.inf)
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                if len(state) == 0:
                    state['step'] = 0
                    state['avg_utility'] = torch.zeros_like(p.data)
                state['step'] += 1
                avg_utility = state['avg_utility']
                avg_utility.mul_(group['beta_utility']).add_(
                    -p.grad.data*p.data,
                    alpha=1-group['beta_utility']
                )
                current_max_util = avg_utility.max()
                if current_max_util > global_max_util:
                    global_max_util = current_max_util
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                bias_correction_util = 1-group['beta_utility']**state['step']
                noise = torch.randn_like(p.grad)*group['sigma']
                scaled_util = torch.sigmoid_((state['avg_utility']/bias_correction_util)/global_max_util)
                p.data.mul_(
                    1-group['lr']*group['weight_decay']
                ).add_(
                    (p.grad.data+noise)*(1-scaled_util),
                    alpha=-2.0*group['lr'],
                )



import torch
import numpy as np

class Memory:
    def __init__(
        self,
        batch_size
    ):
        self.batch_size = batch_size

        self.states = []
        self.actions = []
        self.probs = []
        self.vals = []
        self.rewards = []
        self.dones = []

    def recall(self):
        n_states = len(self.states)
        batch_start = np.arange(0, n_states, self.batch_size)
        indices = np.arange(n_states)
        np.random.shuffle(indices)
        batches = [
            indices[i:i + self.batch_size]
            for i in batch_start
        ]

        actions = [action.to('cpu') for action in self.actions]
        probs = [prob.to('cpu') for prob in self.probs]
        vals = [val.to('cpu') for val in self.vals]

        return self.states, \
            np.array(actions), \
            np.array(probs), \
            np.array(vals), \
            np.array(self.rewards), \
            np.array(self.dones), \
            batches
    
    def store(
        self,
        state,
        action,
        prob,
        val,
        reward,
        done
    ):
        self.states.append(state)
        self.actions.append(action)
        self.probs.append(prob)
        self.vals.append(val)
        self.rewards.append(reward)
        self.dones.append(done)

    def clear(self):
        self.states = []
        self.actions
        self.probs = []
        self.vals = []
        self.rewards = []
        self.dones = []
        torch.cuda.empty_cache()



import warnings

import torch
import torch.nn as nn
import torch.nn.functional as functional

from torch.optim import Optimizer
from torch.distributions.categorical import Categorical

from .transformer import TransformerBlock, AttentionGate

from ..optimizers import Adam, UPGD

from ..utils import generate_layers

class PolicyNetwork(nn.Module):
    def __init__(
        self,
        input_dims: int,
        output_dims: int,
        transformer_layers: int = 2,
        optimizer_name: str = 'adam',
        device='cpu',
        **kwargs
    ):
        gate_hparams = {
            **kwargs.get("gate", {})
        }
        network_defaults = dict(
            layers=None,
            activation_fn=None
        )
        network_architecture = {
            **network_defaults,
            **kwargs.get("network", {})
        }
        transformer_hparams = {
            **kwargs.get("transformer", {})
        }
        super(PolicyNetwork, self).__init__()

        self.transformer = nn.ModuleList(
            [
                TransformerBlock(
                    embed_dims=input_dims,
                    **transformer_hparams,
                )
                for _ in range(transformer_layers)
            ]
        )
        
        self.gate = AttentionGate(
            embed_dims=input_dims,
            **gate_hparams
        )

        if network_architecture['layers'] is not None:
            layers = generate_layers(
                input_dims=input_dims,
                output_dims=output_dims,
                **network_architecture
            )
            self.actor = nn.ModuleList(layers)
        else:
            self.actor = nn.ModuleList(
                [
                    nn.Linear(input_dims, output_dims)
                ]
            )

        self.to(device)

        ## OPTIMIZER
        optimizer = kwargs.get("optimizer")
        if isinstance(optimizer, type) and issubclass(optimizer, Optimizer):
            optimizer_hparams = {
                **kwargs.get("optimizer_hparams", {})
            }
            self.optimizer = optimizer(
                self.parameters(),
                **optimizer_hparams
            )
        else: 
            ####### TO BE REMOVED DO NOT USE THIS
            warnings.warn(
                "Passing 'optimizer' as a dict to the actor/critic is deprecated. "
                "Please pass a pytorch.optim class in 'optimizer' and hyperparameters in 'optimizer_hparams' dict.",
                Warning
            )
            optimizer_hparams = {k: v for k, v in optimizer.items()}
            if 'upgd' in optimizer_name.lower():
                self.optimizer = UPGD(
                    params=self.parameters(),
                    **optimizer_hparams
                )
            else:
                self.optimizer = Adam(
                    params=self.parameters(),
                    **optimizer_hparams
                )
            #####################################


    def forward(self, state, mask=None):
        out = state
        for layer in self.transformer:
            out = layer(out, out, out, mask)
        out = self.gate(out)
        for layer in self.actor:
            out = layer(out)
        out = functional.softmax(out, -1)
        dist = Categorical(out)
        return dist
